/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _ALLELEPOOL_H_
#define _ALLELEPOOL_H_


#include <stdexcept>
#include <string>
#include <map>
#include "util.h"


typedef size_t Allele;


//
// A source for managing unique alleles.
//

class AllelePool {
public:
	AllelePool();

	size_t size() const;
	bool contains(const std::string& value) const;
	bool contains(Allele a) const;
	Allele get(const std::string& value) const throw(std::invalid_argument);
	const std::string& get(Allele a) const throw(std::out_of_range);
	Allele highestAllele() const throw(std::domain_error);

	Allele newAllele(const std::string& value) throw(std::invalid_argument);
 
	bool load(const char* filename);
	friend std::istream& operator>>(std::istream& in, AllelePool& ap);
	bool save(const char* filename) const;
	friend std::ostream& operator<<(std::ostream& out, const AllelePool& sp);

private:
	void set(Allele a, const std::string& value) throw(std::invalid_argument);

	std::map<Allele,std::string> alleleToString;
	std::map<std::string,Allele> stringToAllele;
	Allele highestFreeAllele;
};


inline AllelePool::AllelePool()
: highestFreeAllele(0) {
}

inline size_t AllelePool::size() const {
	return alleleToString.size();
}

inline bool AllelePool::contains(const std::string& value) const {
	return stringToAllele.find(value) != stringToAllele.end();
}

inline bool AllelePool::contains(Allele a) const {
	return alleleToString.find(a) != alleleToString.end();
}


inline bool AllelePool::load(const char* filename) {
	return ::load(filename, this);
}

inline bool AllelePool::save(const char* filename) const {
	return ::save(filename, this);
}

#endif
