/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <cmath>
#include <sstream>
#include "Chi2Approx.h"


using namespace std;


//Chi2Approx g_Chi2Approx("chi2.dat");
Chi2Approx g_Chi2Approx;


Chi2Approx::Chi2Approx(const char* filename) throw(runtime_error) {
	if (!load(filename))
		throw runtime_error(string("Error: Unable to load ") + filename);
}

double Chi2Approx::cdf(unsigned int DoF, double chi2) const throw(out_of_range) {
	if (chi2 < 0) {
		stringstream ss;
		ss << "Chi2Approx::cdf() out_of_range: chi2 = " <<  chi2;
		throw out_of_range(ss.str());
	}

	size_t index;
	double target;

	if (DoF < levels.size()) {
		index = DoF - 1;
		target = chi2;
	} else {
		index = levels.size() - 1;
		target = (chi2 - DoF) / sqrt(static_cast<double>(2 * DoF));
		if (target < 0)
			target = -target;
	}

	double result;
	if (target < levels[index][0].chi2)
		if (index == levels.size() - 1)
			result = 0.5;
		else
			result = 0;
	else if (target >= levels[index][levels[index].size() - 1].chi2)
		result = 1.0;
	else {
		size_t low = 0, high = levels[index].size() - 2;
		while (true) {
			size_t mid = (low + high) / 2;
			if (levels[index][mid].chi2 > target)
				high = mid - 1;
			else if (levels[index][mid+1].chi2 <= target)
				low = mid + 1;
			else {
				assert(levels[index][mid].chi2 <= target && target < levels[index][mid+1].chi2);
				double frac = (target - levels[index][mid].chi2) / (levels[index][mid+1].chi2 - levels[index][mid].chi2);
				result = levels[index][mid].cdf * (1 - frac) + levels[index][mid+1].cdf * frac;
				break;
			}
		}
	}
	if (index == levels.size() - 1 && chi2 < DoF)
		result = 1 - result;
	return result;
}

istream& operator>>(istream& in, Chi2Approx& ca) {
	unsigned int DoFs, points;
	double a, b;

	in >> DoFs;
	ca.levels.resize(DoFs + 1);
	for (unsigned int i = 0; i <= DoFs; i++) {
		in >> points;
		ca.levels[i].resize(points + 1);
		if (i < DoFs) {
			ca.levels[i][0].chi2 = 0;
			ca.levels[i][0].cdf = 0;
		} else {
			ca.levels[i][0].chi2 = 0;
			ca.levels[i][0].cdf = 0.5;
		}
		for (unsigned int j = 1; j <= points; j++) {
			in >> a >> b;
			ca.levels[i][j].chi2 = a;
			ca.levels[i][j].cdf = b;
		}
	}
	return in;
}

#ifdef SELFTEST
#include <gsl/gsl_cdf.h>
#include "DiscreteStatistics.h"
#include "Random.h"

void Chi2Approx::selfTest() {
	Chi2Approx chi2("chi2.dat");

	DoubleStatistics stats;
	for (int dof = 1; dof <= 70; dof++) {
		stats.clear();
		for (int i = 0; i < 1000; i++) {
			double target = g_Random.nextOpen();
			double x = gsl_cdf_chisq_Pinv(target, dof);
			if (dof > 50)
				target = gsl_cdf_ugaussian_P((x - dof) / sqrt(2 * dof));
			double approx = chi2.cdf(dof, x);
//			assert(fabs(approx - target) <= 0.01);
			stats.add(fabs(approx - target));
		}
		cout << "DoF: " << dof << endl;
		cout << "Mean/Median/Max: " << stats.mean() << '/' << stats.median() << '/' << stats.max() << endl;
		assert(stats.max() < 0.01);
	}
}

#endif
