/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _FISWRAPPER_H_
#define _FISWRAPPER_H_


#include <cassert>
#include <sstream>
#include "FiniteIntegerSet.h"


//
// This class defines a FiniteIntegerSet that excludes one value from being
// either a member or non-member
//

class FISWrapper : public FiniteIntegerSet {
public:
	typedef MemberIterator Iterator;

	FISWrapper(unsigned int genes, unsigned int exclude);

	size_t numberOfNonMembers() const;
	bool isMember(unsigned int value) const throw(std::out_of_range);
	unsigned int randomNonMember() const throw(std::domain_error);

	void addMember(unsigned int value) throw(std::out_of_range);

#ifdef SELFTEST
	static void selfTest();
#endif

private:
	unsigned int exclude;
};


inline FISWrapper::FISWrapper(unsigned int genes, unsigned int exclude)
: FiniteIntegerSet(genes - 1), exclude(exclude) {
	assert(genes > 1);
	assert(exclude < genes);
}

inline size_t FISWrapper::numberOfNonMembers() const {
	return FiniteIntegerSet::numberOfNonMembers() - 1;
}

inline bool FISWrapper::isMember(unsigned int value) const throw(std::out_of_range) {
	if (value == exclude) {
		std::stringstream ss;
		ss << "FISWrapper::isMember() out_of_range: Excluded gene cannot be a member: " << value;
		throw std::out_of_range(ss.str());
	}
	return FiniteIntegerSet::isMember(value);
}

inline unsigned int FISWrapper::randomNonMember() const throw(std::domain_error) {
	if (numberOfNonMembers() == 0)
		throw std::domain_error("FISWrapper::randomNonMember() domain_error: There are no non-member integers");

	unsigned int gene;
	while ((gene = FiniteIntegerSet::randomNonMember()) == exclude);
	return gene;
}

inline void FISWrapper::addMember(unsigned int value) throw(std::out_of_range) {
	if (value == exclude) {
		std::stringstream ss;
		ss << "FISWrapper::addMember() out_of_range: Cannot add excluded gene as member: " << value;
		throw std::out_of_range(ss.str());
	}

	FiniteIntegerSet::addMember(value);
}


#endif
