/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <new>
#include <set>
#include <sstream>
#include "Chi2Approx.h"
#include "DiscreteStatistics.h"
#include "Random.h"
#include "MARLEDA.h"


using namespace std;


#define DETAIL_LOG(x) {if (detailLog) *detailLog << x;}
#define FITNESS_LOG(x) {if (fitnessLog) *fitnessLog << x;}
#define REG(x) x(params[#x])
#define REQUIRED(x) if (!ps.get(#x,x)) {cerr << "Error: Required parameter \"" #x "\" missing" << endl; return false;}


MARLEDA::MARLEDA()
: detailLog(NULL), fitnessLog(NULL), task(NULL), generation(0), evaluation(0),
replaceMask(population), selectionMask(population),
saveSnapshots(false), openedDetailLog(false), openedFitnessLog(false),
REG(generations), REG(evaluations), REG(popSize), REG(genes),
REG(mutation), REG(elitists), REG(selectionBasis), REG(selectionSize),
REG(selectionMin),
REG(tournamentSize), REG(modelAddTries), REG(modelAddThresh),
REG(modelSubTries), REG(modelSubThresh), REG(monteIters),
REG(snapshotFreq), REG(snapshotMin), REG(snapshotPrefix) {
	generations = -1;
	evaluations = -1;
	mutation = 0;
	elitists = 0;
	selectionBasis = 1.0;
	selectionSize = 1.0;
	selectionMin = 1;
	snapshotFreq = 1;
	snapshotMin = -1;
	snapshotTimer = 0;
}

bool MARLEDA::init(ParameterSet& ps, EvolutionTask* task) {
	assert(task != NULL);
	this->task = task;

	string tpString;
	if (ps.get("allelePool", tpString)) {
		if (!allelePool.load(tpString.c_str())) {
			cerr << "Error: unable to load allele definitions from \"" << tpString << "\"" << endl;
			return false;
		}
	} else {
		if (!task->initAlleles(ps, allelePool) ||
		    allelePool.size() == 0) {
			cerr << "Error: AllelePool is empty" << endl;
			return false;
		}
	}
	if (allelePool.highestAllele() != allelePool.size() - 1) {
		cerr << "Error: Alleles are not contiguous" << endl;
		return false;
	}
	if (!task->init(ps, allelePool))
		return false;
	alleles = allelePool.size();

	if (!ps.contains("generations") && !ps.contains("evaluations")) {
		cerr << "Error: At least one of parameters \"generations\" and \"evaluations\" is required" << endl;
		return false;
	}
	ps.get("generations", generations);
	ps.get("evaluations", evaluations);
	ps.get("mutation", mutation);
	ps.get("elitists", elitists);
	ps.get("selectionBasis", selectionBasis);
	ps.get("selectionSize", selectionSize);
	ps.get("selectionMin", selectionMin);
	REQUIRED(tournamentSize);
	REQUIRED(modelAddTries);
	REQUIRED(modelAddThresh);
	REQUIRED(modelSubTries);
	REQUIRED(modelSubThresh);
	REQUIRED(monteIters);

	if (ps.get("popFile", tpString)) {
		// Load populatoin
		if(!population.load(tpString.c_str(), allelePool)) {
			cerr << "Error: unable to load population file '" << tpString << "'" << endl;
			return false;
		}
		popSize = population.size();
		genes = population[0].chromosome->size();
	} else {
		REQUIRED(popSize);
		REQUIRED(genes);
		generateRandomPopulation();
	}

	if (ps.get("modelFile", tpString)) {
		if (!model.load(tpString.c_str())) {
			cerr << "Error: unable to load model file '" << tpString << "'" << endl;
			return false;
		}
	} else {
		if (!model.init(genes))
			return false;
	}

	if (ps.get("snapshot", snapshotPrefix)) {
		ps.get("snapshotMin", snapshotMin);
//		if (snapshotMin < 0) {
//			cerr << "Error: minimum snapshot generation must be at least 0" << endl;
//			return false;
//		}
		if (generations > 0 && snapshotMin > generations)
			cerr << "Warning: Minimum snapshot generation is greater than total generations" << endl;

		ps.get("snapshotFreq", snapshotFreq);
		if (snapshotFreq < 1) {
			cerr << "Error: Snapshot frequency must be at least 1" << endl;
			return false;
		}

		string::size_type pos = snapshotPrefix.find_last_of('/');
		if (pos != string::npos) {
			if (!makedir(snapshotPrefix.substr(0, pos).c_str()))
				return false;
			ps.save((snapshotPrefix.substr(0, pos) + "/debug.param").c_str());
		}

		saveSnapshots = true;
	}

	if (ps.get("detailLog", tpString)) {
		if (tpString == "-")
			detailLog = &cout;
		else {
			string path = pathDirectory(tpString.c_str());
			if (!makedir(path.c_str()))
				return false;
			detailLog = new ofstream(tpString.c_str());
			if (!detailLog->good() || dynamic_cast<ofstream*>(detailLog)->is_open() == 0) {
				cerr << "Error: Unable to open log file \"" << tpString << "\"" << endl;
				delete detailLog;
				return false;
			}
			openedDetailLog = true;
		}
	}
	if (ps.get("fitnessLog", tpString)) {
		if (tpString == "-")
			fitnessLog = &cout;
		else {
			string path = pathDirectory(tpString.c_str());
			if (!makedir(path.c_str()))
				return false;
			fitnessLog = new ofstream(tpString.c_str());
			if (!fitnessLog->good() || dynamic_cast<ofstream*>(fitnessLog)->is_open() == 0) {
				cerr << "Error: Unable to open log file \"" << tpString << "\"" << endl;
				delete fitnessLog;
				return false;
			}
			openedFitnessLog = true;
			}
	}

	if (elitists > popSize) {
		cerr << "Error: elistists > popSize" << endl;
		return false;
	}
	if (elitists < 1.0)
		elitists = round(elitists * popSize);
	replaceMask.reserve(static_cast<size_t>(popSize - elitists));
	for (size_t i = static_cast<size_t>(elitists); i < static_cast<size_t>(popSize); i++)
		replaceMask.add(i);
	children.resize(replaceMask.size());

	if (selectionBasis > popSize) {
		cerr << "Error: selectionBasis > population" << endl;
		return false;
	}
	if (selectionSize > popSize && tournamentSize < 2) {
		cerr << "Error: selectionSize > popSize when not using tournament selection" << endl;
		return false;
	}
	if (selectionBasis <= 1.0)
		selectionBasis = round(selectionBasis * popSize);
	if (selectionSize <= 1.0)
		selectionSize = round(selectionSize * popSize);
	if (tournamentSize < 2)
		selectionSize = popSize;
	// Effectively remove selectionSize parameter for the time being
	selectionSize = popSize;
	selectionMask.reserve(static_cast<size_t>(selectionSize));
	if (tournamentSize < 2) {
		for (size_t i = 0; i < static_cast<size_t>(selectionSize); i++)
			selectionMask.add(i);
	}
	if (selectionMin < 1.0)
		selectionMin = round(selectionMin * selectionSize);
	if (selectionMin < 1) {
		cerr << "Error: selectionMin must be positive" << endl;
		return false;
	}
	
	allelePDFCache.resize(genes);
	marginalPDFs.resize(genes);
	for (size_t i = 0; i < static_cast<size_t>(genes); i++)
		marginalPDFs[i] = new AllelePDF(alleles);

	return true;
}

MARLEDA::~MARLEDA() {
	for (size_t i = 0; i < marginalPDFs.size(); i++)
		delete marginalPDFs[i];

	if (openedDetailLog) {
		detailLog->clear();
		dynamic_cast<ofstream*>(detailLog)->close();
		delete dynamic_cast<ofstream*>(detailLog);
	}
	if (openedFitnessLog) {
		fitnessLog->clear();
		dynamic_cast<ofstream*>(fitnessLog)->close();
		delete dynamic_cast<ofstream*>(fitnessLog);
	}
}

void MARLEDA::evolve() {
	population.sort();
	if (saveSnapshots && snapshotMin >= 0 && generation >= snapshotMin) {
		if (!snapshot()) {
			cerr << "Error: Unable to save snapshot" << endl;
			return;
		}
		snapshotTimer = snapshotFreq;
	}
	DoubleStatistics fitnessStats;
	for (size_t i = 0; i < population.size(); i++)
		fitnessStats.add(population[i].fitness);
	FITNESS_LOG("Generation\tEvaluations\tMin\tMean\tMedian\tMax\tStdDev" << endl);
	FITNESS_LOG(generation << "\t" << evaluation << "\t" << fitnessStats.min() << "\t" << fitnessStats.mean() << "\t" << fitnessStats.median() << "\t" << fitnessStats.max() << "\t" << fitnessStats.stddev() << endl);
//	FITNESS_LOG("Generation\tEvaluations\tFitness" << endl);
//	FITNESS_LOG(generation << "\t" << evaluation << "\t" << population[0].fitness << endl);

	while ((generations < 0 || generation < generations) && 
	       (evaluations < 0 || evaluation < evaluations)) {
		generation++;
		if (snapshotTimer > 0)
			snapshotTimer--;
		DETAIL_LOG("Generation " << generation << endl);
		runTournament();
		calculateMarginals();
		growModel();
		for (size_t i = 0; i < static_cast<size_t>(genes); i++)
			allelePDFCache[i].clear();
#ifndef THREADED
		for (size_t i = 0; i < children.size(); i++)
			children[i] = createChromosome();
#else
		manageCreationThreads();
#endif
		for (size_t i = 0; i < replaceMask.size(); i++) {
			delete replaceMask[i].chromosome;
			replaceMask[i].chromosome = children[i];
		}
		task->evaluate(replaceMask);
		evaluation += replaceMask.size();
		population.sort();
		DETAIL_LOG(endl << endl);

		if (saveSnapshots && snapshotMin >= 0 && (generation == snapshotMin || snapshotTimer == 0)) {
			if (!snapshot()) {
				cerr << " Error: Unable to save snapshot" << endl;
				return;
			}
			snapshotTimer = snapshotFreq;
		}
		fitnessStats.clear();
		for (size_t i = 0; i < population.size(); i++)
			fitnessStats.add(population[i].fitness);
		FITNESS_LOG(generation << "\t" << evaluation << "\t" << fitnessStats.min() << "\t" << fitnessStats.mean() << "\t" << fitnessStats.median() << "\t" << fitnessStats.max() << "\t" << fitnessStats.stddev() << endl);
//		FITNESS_LOG(generation << "\t" << evaluation << "\t" << population[0].fitness << endl);
	}
	if (saveSnapshots && (snapshotMin < 0 || snapshotTimer != snapshotFreq)) {
		if (!snapshot()) {
			cerr << " Error: Unable to save snapshot" << endl;
			return;
		}
	}
}

// TODO: add efficient processing when neighborhood is empty
// TODO: re-enable Yates' correction
double MARLEDA::confidenceAfterLink(Gene gene1, Gene gene2) {
	VectorMap<Allele,vector<size_t> > cache;

	vector<Gene> neighbors(model.numberOfNeighbors(gene1));
	model.getNeighbors(gene1, neighbors);

	// Build the cache of neighborhood allele combinations
	vector<Allele> key(model.numberOfNeighbors(gene1));
	for (size_t i = 0; i < selectionMask.size(); i++) {
		for (size_t j = 0; j < neighbors.size(); j++)
			key[j] = (*selectionMask[i].chromosome)[neighbors[j]];
		cache.get(key).push_back(i);
	}

	vector<size_t> gene1PDF(alleles), gene2PDF(alleles), jointPDF(alleles * alleles);
	double chi2 = 0;

	VectorMap<Allele,vector<size_t> >::iterator it = cache.begin();
	while (it != cache.end()) {
		for (size_t i = 0; i < alleles; i++)
			gene1PDF[i] = gene2PDF[i] = 0;
		for (size_t i = 0; i < alleles * alleles; i++)
			jointPDF[i] = 0;

		vector<size_t> &chrms = it->second;
		it++;
		for (size_t i = 0; i < chrms.size(); i++) {
			Allele a1 = (*selectionMask[chrms[i]].chromosome)[gene1];
			Allele a2 = (*selectionMask[chrms[i]].chromosome)[gene2];
			jointPDF[a1 * alleles + a2]++;
		}
		for (Allele i = 0; i < alleles; i++)
			for (Allele j = 0; j < alleles; j++) {
				gene1PDF[i] += jointPDF[i * alleles + j];
				gene2PDF[j] += jointPDF[i * alleles + j];
			}

		for (Allele i = 0; i < alleles; i++)
			for (Allele j = 0; j < alleles; j++) {
				double expected = static_cast<double>(gene1PDF[i] * gene2PDF[j]) / chrms.size();
				double tmp = jointPDF[i * alleles + j] - expected;
//				if (alleles == 2) // Yates' correction
//					tmp = fabs(tmp) - 0.5;
				if (expected > 0)
					chi2 += tmp * tmp / expected;
				else if (jointPDF[i * alleles + j] > 0)
					chi2 += jointPDF[i * alleles + j];
			}
	}

	double oldDoF = 1;
	size_t allelesPresent = 0;
	for (size_t j = 0; j < alleles; j++)
		if (marginalPDFs[gene1]->frequency(j) > 0)
			allelesPresent++;
	if (allelesPresent >= 2)
		oldDoF *= allelesPresent - 1;
//	if (allelesPresent >= 3)
//		oldDoF *= allelesPresent - 1;
//	else if (allelesPresent == 2)
//		oldDoF *= 1.75;
	allelesPresent = 0;
	for (size_t j = 0; j < alleles; j++)
		if (marginalPDFs[gene2]->frequency(j) > 0)
			allelesPresent++;
	if (allelesPresent >= 2)
		oldDoF *= allelesPresent - 1;
//	if (allelesPresent >= 3)
//		oldDoF *= allelesPresent - 1;
//	else if (allelesPresent == 2)
//		oldDoF *= 1.75;
	double newDoF = oldDoF;
	for (size_t i = 0; i < neighbors.size(); i++) {
		allelesPresent = 0;
		for (size_t j = 0; j < alleles; j++)
			if (marginalPDFs[i]->frequency(j) > 0)
				allelesPresent++;
		if (allelesPresent >= 3)
			oldDoF *= allelesPresent - 1;
		else if (allelesPresent == 2)
			oldDoF *= 1.75;
	}

	if (cache.size() > 1)
		newDoF *= cache.size() - 1;

	return g_Chi2Approx.cdf(static_cast<size_t>(newDoF), chi2);
}

double MARLEDA::confidenceAfterUnlink(Gene gene1, Gene gene2) {
	model.removeNeighbor(gene1, gene2);
	double retVal = confidenceAfterLink(gene1, gene2);
	model.addNeighbor(gene1, gene2);
	return retVal;
}

// TODO: don't duplicate confidence checks when the neighborhood hasn't changed
void MARLEDA::growModel() {
	size_t sAddTries = 0, sAddSucc = 0, sSubTries = 0, sSubSucc = 0;
	size_t addTriesLeft = modelAddTries, subTriesLeft = modelSubTries;
	while (true) {
		bool addition;
		if (addTriesLeft + subTriesLeft == 0)
			break;
		else if (addTriesLeft == 0 && model.numberOfNotEmpties() == 0)
			break;
		else if (subTriesLeft == 0 && model.numberOfNotFulls() == 0)
			break;
		else if (model.numberOfNotEmpties() == 0)
			addition = true;
		else if (model.numberOfNotFulls() == 0)
			addition = false;
		else
			addition = g_Random.next32(0, addTriesLeft + subTriesLeft - 1) < addTriesLeft;
		if (addition) {	
			assert(addTriesLeft > 0);
			sAddTries++;
			Gene gene1 = model.randomNotFull();
			Gene gene2 = model.randomNonNeighbor(gene1);
			double tmp = confidenceAfterLink(gene1, gene2);
			if (tmp >= modelAddThresh) {
				model.addNeighbor(gene1, gene2);
				sAddSucc++;
			}
			addTriesLeft--;
		} else {
			assert(subTriesLeft > 0);
			sSubTries++;
			Gene gene1 = model.randomNotEmpty();
			Gene gene2 = model.randomNeighbor(gene1);
			double tmp = confidenceAfterUnlink(gene1, gene2);
			if (tmp < modelSubThresh) {
				model.removeNeighbor(gene1, gene2);
				sSubSucc++;
			}
			subTriesLeft--;
		}
	}
	if (sAddTries > 0)
		DETAIL_LOG(sAddSucc << "/" << sAddTries << " (" << static_cast<int>(100 * static_cast<double>(sAddSucc) / sAddTries) << "%) model additions successful" << endl);
	if (sSubTries > 0)
		DETAIL_LOG(sSubSucc << "/" << sSubTries << " (" << static_cast<int>(100 * static_cast<double>(sSubSucc) / sSubTries) << "%) model subtractions successful" << endl);
	if (detailLog) {
		size_t sum = 0;
		for (size_t i = 0; i < static_cast<size_t>(genes); i++)
			sum += model.numberOfNeighbors(i);
		*detailLog << "Total model neighbors: " << sum/2 << endl;
	}
}

void MARLEDA::runTournament() {
	// Run tournament to bias selectionMask
	if (tournamentSize >= 2) {
		selectionMask.clear();
		for (size_t i = 0; i < static_cast<size_t>(selectionSize); i++) {
			size_t best = g_Random.next32(0, static_cast<size_t>(selectionBasis) - 1);
			for (size_t j = 1; j < static_cast<size_t>(tournamentSize); j++) {
				size_t opp = g_Random.next32(0, static_cast<size_t>(selectionBasis) - 1);
				if (population[opp].fitness > population[best].fitness)
					best = opp;

			}
			selectionMask.add(best);
		}
	}
}

// TODO: only sample genes with no neighbors once
#ifdef THREADED
#define LOW lowGene
#define HIGH highGene
#define ITERS monteIters
#define RANDOM (*random)
void MARLEDA::createChromosomes(size_t lowGene, size_t highGene, size_t monteIters, Random* random) {
	vector<Gene> neighbors;
	vector<Allele> key;

	for (size_t index = 0; index < children.size(); index++) {
	Chromosome *c = children[index];
#else
#define LOW 0
#define HIGH genes - 1
#define ITERS static_cast<size_t>(monteIters)
#define RANDOM g_Random
Chromosome* MARLEDA::createChromosome() {
	vector<Gene> neighbors;
	vector<Allele> key;

	size_t original = RANDOM.next32(0, selectionMask.size()-1);
	Chromosome *c = new Chromosome(*selectionMask[original].chromosome);
#endif

	for (size_t i = 0; i < ITERS; i++) {
		Gene gene = RANDOM.next32(LOW, HIGH);
		if (model.numberOfNeighbors(gene) == 0) {
#ifdef THREADED
			(*c)[gene] = marginalPDFs[gene]->sample(&RANDOM);
#else
			(*c)[gene] = marginalPDFs[gene]->sample();
#endif
			continue;
		}
		model.getNeighbors(gene, neighbors);
		key.resize(neighbors.size());
		for (size_t j = 0; j < neighbors.size(); j++)
			key[j] = (*c)[neighbors[j]];

		VectorMap<Allele,AllelePDF>::iterator it = allelePDFCache[gene].find(key);
		if (it == allelePDFCache[gene].end())
			it = allelePDFCache[gene].insert(make_pair(key, AllelePDF(alleles))).first;
		AllelePDF &apdf = it->second;
		if (apdf.totalFrequency() == 0) {
			for (size_t m = 0; m < neighbors.size() && apdf.totalFrequency() < selectionMin; m++) {
				apdf.clear();
				for (size_t j = 0; j < selectionMask.size(); j++) {
					size_t k, diff = 0;
					for (k = 0; k < neighbors.size(); k++)
						if ((*selectionMask[j].chromosome)[neighbors[k]] != (*c)[neighbors[k]]) {
							diff++;
							if (diff > m)
								break;
						}
					if (k == neighbors.size())
						apdf.add((*selectionMask[j].chromosome)[gene]);
				}
				apdf.prepareForSampling();
			}
			if (apdf.totalFrequency() < selectionMin) {
				apdf.clear();
				apdf = *marginalPDFs[gene];
			}
		}
#ifdef THREADED
		(*c)[gene] = apdf.sample(&RANDOM);
#else
		(*c)[gene] = apdf.sample();
#endif
	}

	if (mutation > 0) {
		for (Gene i = LOW; i < HIGH + 1; i++)
			if (RANDOM.nextHalf() < mutation) {
				Allele a = RANDOM.next32(0, alleles - 2);
				if (a != (*c)[i])
					(*c)[i] = a;
				else
					(*c)[i] = alleles - 1;
			}
	}

#ifdef THREADED
	}
#else
	return c;
#endif
}

void MARLEDA::calculateMarginals() {
	for (size_t i = 0; i < static_cast<size_t>(genes); i++)
		marginalPDFs[i]->clear();
	for (size_t i = 0; i < selectionMask.size(); i++)
		for (size_t j = 0; j < static_cast<size_t>(genes); j++) 
			marginalPDFs[j]->add((*selectionMask[i].chromosome)[j]);
	for (size_t i = 0; i < static_cast<size_t>(genes); i++)
		marginalPDFs[i]->prepareForSampling();
}

void MARLEDA::generateRandomPopulation() {
	population.resize(popSize);
	for (size_t i = 0; i < static_cast<size_t>(popSize); i++) {
		population[i].chromosome = new Chromosome(genes);
		for (Gene j = 0; j < static_cast<Gene>(genes); j++)
			(*population[i].chromosome)[j] = g_Random.next32(0,alleles-1);
	}
	PopulationMask all(population, true);
	task->evaluate(all);
	evaluation += all.size();
}

bool MARLEDA::snapshot() const {
	stringstream ss;
	ss << snapshotPrefix << generation << ".pop";
	if (!population.save(ss.str().c_str(), allelePool))
		return false;
	ss.str("");
	ss << snapshotPrefix << generation << ".model";
	if (!model.save(ss.str().c_str()))
		return false;
	return true;
}

#ifdef THREADED
struct CreationThreadArgs {
	MARLEDA *marleda;
	size_t lowGene, highGene, monteIters;
	Random *random;
};

void MARLEDA::manageCreationThreads() {
	// Create all initial chromosomes
	for (size_t i = 0; i < children.size(); i++) {
		size_t original = g_Random.next32(0, selectionMask.size()-1);
		children[i] = new Chromosome(*selectionMask[original].chromosome);
	}

	// Make sure all new threads are joinable
	pthread_attr_t threadAttribs;
	pthread_attr_init(&threadAttribs);
	pthread_attr_setdetachstate(&threadAttribs, PTHREAD_CREATE_JOINABLE);

	// Set up thread structures
	size_t threads = sysconf(_SC_NPROCESSORS_ONLN);
	vector<pthread_t> newThreads(threads-1);
	// Create a unique RNG for each thread so we don't have to
	// guard the global RNG with a slow Linux pthread mutex
	vector<Random> newRNGs(threads-1);
	vector<CreationThreadArgs> threadArgs(threads);
	threadArgs[0].marleda = this;
	threadArgs[0].lowGene = 0;
	threadArgs[0].highGene = genes / threads - 1;
	threadArgs[0].monteIters = (threadArgs[0].highGene + 1) * monteIters / genes;
	threadArgs[0].random = &g_Random;
	for (size_t i = 1; i < threads; i++) {
		threadArgs[i].marleda = this;
		threadArgs[i].lowGene = threadArgs[i-1].highGene + 1;
		threadArgs[i].highGene = threadArgs[i].lowGene + threadArgs[0].highGene;
		threadArgs[i].monteIters = threadArgs[0].monteIters;
		newRNGs[i-1].reseed(g_Random.next32());
		threadArgs[i].random = &newRNGs[i-1];
	}
	threadArgs.back().highGene = genes - 1;
	threadArgs.back().monteIters = monteIters - (threadArgs[0].monteIters * (threads - 1));

	// Spawn new threads
	for (size_t i = 0; i < newThreads.size(); i++) {
		int status = pthread_create(&newThreads[i], &threadAttribs, createChromosomes, &threadArgs[i]);
		if (status != 0) {
			cerr << "Error: unable to create additional threads" << endl;
			return;
		}
	}
	createChromosomes(&threadArgs.back());

	// Wait for all threads to finish
	for (size_t i = 0; i < newThreads.size(); i++)
		pthread_join(newThreads[i], NULL);
	pthread_attr_destroy(&threadAttribs);
}

void* MARLEDA::createChromosomes(void* args) {
	CreationThreadArgs *trueArgs = static_cast<CreationThreadArgs*>(args);
	trueArgs->marleda->createChromosomes(trueArgs->lowGene, trueArgs->highGene, trueArgs->monteIters, trueArgs->random);

	return NULL;
}
#endif
