/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _MARLEDA_H_
#define _MARLEDA_H_


#ifdef THREADED
#include <pthread.h>
#include <set>
#endif
#include "AllelePDF.h"
#include "EvolutionTask.h"
#include "Model.h"
#include "VectorMap.h"


class MARLEDA {
public:
	MARLEDA();
	// init should only be called once per MARLEDA object
	bool init(ParameterSet& ps, EvolutionTask* task);
	~MARLEDA();

	void evolve();

	std::ostream *detailLog, *fitnessLog;

private:
	double confidenceAfterLink(Gene gene1, Gene gene2);
	double confidenceAfterUnlink(Gene gene1, Gene gene2);
	void growModel();
	void runTournament();
#ifndef THREADED
	Chromosome* createChromosome();
#else
	void createChromosomes(size_t geneLow, size_t geneHigh, size_t monteIters, Random* random);
#endif
	void calculateMarginals();
	void generateRandomPopulation();
	bool snapshot() const;
#ifdef THREADED
	void manageCreationThreads();
	static void* createChromosomes(void* args);
#endif
	
	Model model;
	AllelePool allelePool;
	Population population;
	EvolutionTask *task;
	long generation, evaluation;
	PopulationMask replaceMask, selectionMask;
	std::vector<Chromosome*> children;
	std::vector<VectorMap<Allele,AllelePDF> > allelePDFCache;
	std::vector<AllelePDF*> marginalPDFs;
	size_t alleles;
	long snapshotTimer;
	bool saveSnapshots, openedDetailLog, openedFitnessLog;

	// Parameters
	ParameterSet params;
	// Maximum generations of evolution
	long &generations;
	// Maximum function evaluations
	long &evaluations;
	// Population size
	long &popSize;
	// Chromosome length
	long &genes;
	// Mutation rate
	double &mutation;
	// Elitist members ( < 1 proportion  >= 1 absolute)
	double &elitists;
	// Top chromosomes contributing to allele selection ( <= 1 proportion > 1 absolute)
	double &selectionBasis;
	// Size of virtual population after selection ( <= 1 proportion > 1 absolute)
	double &selectionSize;
	// Minimum basis chromosomes ( < 1 proportion >= 1 absolute)
	double &selectionMin;
	// Number of chromosomes that compete in a single tournament
	long &tournamentSize;
	// New model neighbor attempts
	long &modelAddTries;
	// Minimum neighbor confidence
	double &modelAddThresh;
	// Model neighbor removal attempts
	long &modelSubTries;
	// Minimum neighbor confidence for survival
	double &modelSubThresh;
	// Monte Carlo itereations
	long &monteIters;

	// Reporting parameters
	// Number of generations between snapshots
	long &snapshotFreq;
	// First snapshot generation
	long &snapshotMin;
	// Snapshot filename prefix (dir + filename prefix)
	std::string &snapshotPrefix;
};


#endif
