/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <iostream>
#include <new>
#ifndef NDEBUG
#include "Random.h"
#endif
#include "Model.h"


using namespace std;


bool Model::init(Gene genes) {
	assert(genes > 1);
	if (genes <= 1) {
		cerr << "Error: Model class requires at least 2 genes" << endl;
		return false;
	}

	for (size_t i = 0; i < neighborhoods.size(); i++)
		delete neighborhoods[i];
	neighborhoods.resize(genes);
	notEmpties.remax(genes-1);
	notFulls.remax(genes-1);
	for (size_t i = 0; i < genes; i++) {
		neighborhoods[i] = new FISWrapper(genes, i);
		notFulls.addMember(i);
	}
	
	return true;
}


Model::~Model()
{
	for (size_t i = 0; i < neighborhoods.size(); i++)
		delete neighborhoods[i];
}

void Model::getNeighbors(Gene gene, vector<Gene>& neighbors) const throw(out_of_range) {
	FISWrapper::Iterator it = neighborhoods.at(gene)->iterator();
	neighbors.clear();
	neighbors.reserve(neighborhoods[gene]->numberOfMembers() + 1);
	neighbors.resize(neighborhoods[gene]->numberOfMembers());
	for (size_t i = 0; i < neighbors.size(); i++)
		neighbors[i] = *++it;
}

void Model::addNeighbor(Gene gene1, Gene gene2) throw(out_of_range) {
	if (gene1 >= neighborhoods.size() || gene2 >= neighborhoods.size() || gene1 == gene2)
		throw out_of_range("Model::makeNeighbors() out_of_range");

	if (neighborhoods[gene1]->numberOfMembers() == 0)
		notEmpties.addMember(gene1);
	neighborhoods[gene1]->addMember(gene2);
	if (neighborhoods[gene1]->numberOfNonMembers() == 0)
		notFulls.removeMember(gene1);

	if (neighborhoods[gene2]->numberOfMembers() == 0)
		notEmpties.addMember(gene2);
	neighborhoods[gene2]->addMember(gene1);
	if (neighborhoods[gene2]->numberOfNonMembers() == 0)
		notFulls.removeMember(gene2);
}

void Model::removeNeighbor(Gene gene1, Gene gene2) throw(out_of_range) {
	if (gene1 >= neighborhoods.size() || gene2 >= neighborhoods.size() || gene1 == gene2)
		throw out_of_range("Model::makeNeighbors() out_of_range");

	if (neighborhoods[gene1]->numberOfNonMembers() == 0)
		notFulls.addMember(gene1);
	neighborhoods[gene1]->removeMember(gene2);
	if (neighborhoods[gene1]->numberOfMembers() == 0)
		notEmpties.removeMember(gene1);

	if (neighborhoods[gene2]->numberOfNonMembers() == 0)
		notFulls.addMember(gene2);
	neighborhoods[gene2]->removeMember(gene1);
	if (neighborhoods[gene2]->numberOfMembers() == 0)
		notEmpties.removeMember(gene2);
}

istream& operator>>(istream& in, Model& m) {
	unsigned int genes;
	in >> genes >> ws >> detecteof;

	for (size_t i = 0; i < m.neighborhoods.size(); i++)
		delete m.neighborhoods[i];
	m.neighborhoods.resize(genes);
	for (size_t i = 0; i < genes; i++)
		m.neighborhoods[i] = new FISWrapper(genes, i);
	m.notFulls.remax(genes-1);
	m.notEmpties.remax(genes-1);

	stringstream ss;
	for (size_t i = 0; i < genes; i++) {
		ss.clear();
		ss.str("");
		getline(in, ss);
		ss >> ws;
		while (!ss.eof()) {
			unsigned int value;
			ss >> value >> ws;
			try {
				m.neighborhoods[i]->addMember(value);
			} catch (out_of_range& e) {
				in.setstate(ios::failbit);
				return in;
			}
		}
		if (m.neighborhoods[i]->numberOfMembers() > 0)
			m.notEmpties.addMember(i);
		if (m.neighborhoods[i]->numberOfNonMembers() > 0)
			m.notFulls.addMember(i);
	}

	// Check the model for consistency
	for (size_t i = 0; i < genes && !in.fail(); i++) {
		FISWrapper::Iterator it = m.neighborhoods[i]->iterator();
		while (it.hasNext())
			if (!m.neighborhoods[*++it]->isMember(i)) {
				in.setstate(ios::failbit);
				break;
			}
	}

	return in; 
}

ostream& operator<<(ostream& out, const Model& m) {
	out << m.neighborhoods.size() << endl;
	for (size_t i = 0; i < m.neighborhoods.size(); i++) {
		FISWrapper::Iterator it = m.neighborhoods[i]->iterator();
		while (it.hasNext())
			out << *++it << ' ';
		out << endl;
	}
	return out;
}

#ifdef SELFTEST
bool Model::isValid() const {
	for (size_t i = 0; i < neighborhoods.size(); i++) {
		assert(neighborhoods[i]->isValid());
		FISWrapper::Iterator it = neighborhoods[i]->iterator();
		while (it.hasNext())
			assert(neighborhoods[*++it]->isMember(i));
		
		assert(notEmpties.isMember(i) == (neighborhoods[i]->numberOfMembers() != 0));
		assert(notFulls.isMember(i) == (neighborhoods[i]->numberOfNonMembers() != 0));
	}
	assert(notEmpties.isValid());
	assert(notFulls.isValid());

	return true;
}

void Model::selfTest() {
	Model model;
	assert(model.init(10));
	assert(model.isValid());

	vector<Gene> neighbors;
	for (int i = 0; i < 10; i++) {
		assert(model.numberOfNeighbors(i) == 0);
		assert(model.numberOfNonNeighbors(i) == 9);
		model.getNeighbors(i, neighbors);
		assert(neighbors.size() == 0);
	}
	assert(model.numberOfNotEmpties() == 0);
	assert(model.numberOfNotFulls() == 10);
	
	bool exceptionThrown = false;
	try {
		model.randomNotEmpty();
	} catch (domain_error& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);

	exceptionThrown = false;
	try {
		model.addNeighbor(5, 5);
	} catch (out_of_range& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);

	for (int i = 0; i < 5; i++)
		for (int j = 0; j < i; j++)
			model.addNeighbor(i, j);
	assert(model.isValid());
	assert(model.numberOfNotEmpties() == 5);
	assert(model.numberOfNotFulls() == 10);

	for (int i = 5; i < 10; i++)
		for (int j = 0; j < 5; j++)
			model.addNeighbor(i, j);
	assert(model.isValid());
	assert(model.numberOfNotEmpties() == 10);
	assert(model.numberOfNotFulls() == 5);

	for (int i = 5; i < 10; i++)
		for (int j = 5; j < i; j++)
			model.addNeighbor(i, j);
	assert(model.isValid());
	assert(model.numberOfNotEmpties() == 10);
	assert(model.numberOfNotFulls() == 0);

	exceptionThrown = false;
	try {
		model.randomNotFull();
	} catch (domain_error& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);

	for (int i = 0; i < 20; i++) {
		unsigned int dead = model.randomNotEmpty();
		model.removeNeighbor(dead, model.randomNeighbor(dead));
	}
	while (model.numberOfNeighbors(5) > 0)
		model.removeNeighbor(5, model.randomNeighbor(5));

	assert(model.save("selftest.model"));
	model.init(2);
	assert(model.load("selftest.model"));
	assert(model.isValid());
}

#endif
