/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _MODEL_H_
#define _MODEL_H_


#include "Chromosome.h"
#include "FISWrapper.h"


class Model {
public:
	Model();
	bool init(Gene genes);
	~Model();

	size_t numberOfNeighbors(Gene gene) const throw(std::out_of_range);
	size_t numberOfNonNeighbors(Gene gene) const throw(std::out_of_range);
	void getNeighbors(Gene gene, std::vector<Gene>& neighbors) const throw(std::out_of_range);
	Gene randomNeighbor(Gene gene) const throw(std::domain_error, std::out_of_range);
	Gene randomNonNeighbor(Gene gene) const throw(std::domain_error, std::out_of_range);
	size_t numberOfNotEmpties() const;
	size_t numberOfNotFulls() const;
	Gene randomNotEmpty() const throw(std::domain_error);
	Gene randomNotFull() const throw(std::domain_error);

	// Flips (non)neighbor status
	void addNeighbor(Gene gene1, Gene gene2) throw(std::out_of_range);
	void removeNeighbor(Gene gene1, Gene gene2) throw(std::out_of_range);

	bool load(const char* filename) {return ::load(filename, this);}
	friend std::istream& operator>>(std::istream& in, Model& m);
	bool save(const char* filename) const {return ::save(filename, this);}
	friend std::ostream& operator<<(std::ostream& out, const Model& m);

#ifdef SELFTEST
	bool isValid() const;
	static void selfTest();
#endif

private:
	std::vector<FISWrapper*> neighborhoods;
	FiniteIntegerSet notEmpties, notFulls;
};


inline Model::Model()
: notEmpties(0), notFulls(0) {
}

inline size_t Model::numberOfNeighbors(Gene gene) const throw(std::out_of_range) {
	return neighborhoods.at(gene)->numberOfMembers();
}

inline size_t Model::numberOfNonNeighbors(Gene gene) const throw(std::out_of_range) {
	return neighborhoods.at(gene)->numberOfNonMembers();
}

inline Gene Model::randomNeighbor(Gene gene) const throw(std::domain_error, std::out_of_range) {
	return neighborhoods.at(gene)->randomMember();
}

inline Gene Model::randomNonNeighbor(Gene gene) const throw(std::domain_error, std::out_of_range) {
	return neighborhoods.at(gene)->randomNonMember();
}

inline size_t Model::numberOfNotEmpties() const {
	return notEmpties.numberOfMembers();
}

inline size_t Model::numberOfNotFulls() const {
	return notFulls.numberOfMembers();
}

inline Gene Model::randomNotEmpty() const throw(std::domain_error) {
	return notEmpties.randomMember();
}

inline Gene Model::randomNotFull() const throw(std::domain_error) {
	return notFulls.randomMember();
}


#endif
