/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <new>
#ifdef USE_PLOTTER
#include <plotter.h>
#endif
#include <sstream>
#include "SpinGlass.h"


using namespace std;


bool SpinGlass::init(ParameterSet& ps, const AllelePool& ap) {
	if (!BinaryEvolutionTask::init(ps, ap))
		return false;

	string file;
	if (!ps.get("spinFile", file)) {
		cerr << "Error: required parameter 'spinFile' missing" << endl;
		return false;
	}

	ifstream in(file.c_str());
	if (!in || !in.is_open()) {
		cerr << "Error: unable to open \"" << file << "\"" << endl;
		return false;
	}

	width = height = -1;
	in >> width >> height >> endl >> detecteof;
	if (width < 1 || height < 1) {
		cerr << "Error: invalid width or height" << endl;
		in.close();
		return false;
	}
	couplings.clear();
	couplings.reserve(2 * width * height);
	stringstream ss;
	coupling tmp = {0, -1, 0};
	while (in.good()) {
		ss.clear();
		ss.str("");
		getline(in, ss);
		ss >> detecteof;
		while (ss.good()) {
			tmp.spin2 = -1;
			tmp.strength = 0;
			ss >> tmp.spin2 >> tmp.strength >> ws >> detecteof;
			if (tmp.spin2 < 0 || tmp.spin2 >= width * height || tmp.strength == 0) {
				cerr << "Error: invalid coupling (" << tmp.spin2 << ", " << tmp.strength << ") for spin " << tmp.spin1 << endl;
				in.close();
				return false;
			}
			couplings.push_back(tmp);
			if (tmp.spin2 < tmp.spin1)
				swap(couplings.back().spin1, couplings.back().spin2);
		}
		tmp.spin1++;
	}
	in.close();
	ps["genes"] = static_cast<long>(width * height);
	return true;
}

double SpinGlass::evaluate(const Chromosome& c) {
	double sum = 0;
	for (size_t i = 0; i < couplings.size(); i++)
		sum += couplings[i].strength * (c[couplings[i].spin1] == ON ? 1 : -1) * (c[couplings[i].spin2] == ON ? 1 : -1);
	return sum;
}

#ifdef USE_PLOTTER
bool SpinGlass::plot(const char* filename) {
	ofstream out(filename);
	if (!out || !out.is_open()) {
		cerr << "Error: unable to open file '" << filename << "'" << endl;
		return false;
	}

	PSPlotter *plotter = new PSPlotter(cin, out, cerr);
	plotter->openpl();
	plotter->space(0,0,max(static_cast<long>(width),static_cast<long>(height)),max(static_cast<long>(width),static_cast<long>(height)));
	plotter->flinewidth(0.1);
	for (size_t i = 0; i < couplings.size(); i++) {
		int spin1y = height - 1 - couplings[i].spin1 / width;
		int spin1x = couplings[i].spin1 % width;
		assert(spin1x < width && spin1y < height);
		int spin2y = height - 1 - couplings[i].spin2 / width;
		int spin2x = couplings[i].spin2 % width;
		assert(spin2x < width && spin2y < height);
		if (couplings[i].strength > 0)
			plotter->color(0, 0, 0);
		else
			plotter->color(45000, 45000, 45000);
		if (spin1y == spin2y && spin1x + 1 < spin2x) {
			plotter->fline(spin1x, spin1y, spin1x - 0.5, spin1y);
			plotter->fline(spin2x, spin2y, spin2x + 0.5, spin2y);
		} else if (spin1x == spin2x && spin2y + 1 < spin1y) {
			plotter->fline(spin1x, spin1y, spin1x, spin1y - 0.5);
			plotter->fline(spin2x, spin2y, spin2x, spin2y + 0.5);
		} else
			plotter->line(spin1x, spin1y, spin2x, spin2y);
	}
	plotter->linewidth(-1);
	plotter->color(0, 0, 0);
	plotter->filltype(1);
	plotter->fillcolor(65535,65535,65535);
	plotter->ffontsize(0.4);
	stringstream ss;
	for (size_t y = 0; y < height; y++)
		for (size_t x = 0; x < width; x++) {
			plotter->fcircle(x, y, 0.35);
			ss.clear();
			ss.str("");
			ss << "\\f2s\\f1\\sb" << x + (height - 1 - y) * width + 1 << "\\eb";
			plotter->alabel('c', 'c', ss.str().c_str());
		}
	plotter->closepl();
	delete plotter;
	out.close();
	return true;
}

bool SpinGlass::plot(const Chromosome& c, const char* filename) {
	ofstream out(filename);
	if (!out || !out.is_open()) {
		cerr << "Error: unable to open file '" << filename << "'" << endl;
		return false;
	}

	PSPlotter *plotter = new PSPlotter(cin, out, cerr);
	plotter->openpl();
	plotter->space(0,0,max(static_cast<long>(width),static_cast<long>(height)),max(static_cast<long>(width),static_cast<long>(height)));
	plotter->flinewidth(0.1);
	for (size_t i = 0; i < couplings.size(); i++) {
		int spin1y = height - 1 - couplings[i].spin1 / width;
		int spin1x = couplings[i].spin1 % width;
		assert(spin1x < width && spin1y < height);
		int spin2y = height - 1 - couplings[i].spin2 / width;
		int spin2x = couplings[i].spin2 % width;
		assert(spin2x < width && spin2y < height);
		if (couplings[i].strength * (c[couplings[i].spin1] == ON ? 1 : -1) * (c[couplings[i].spin2] == ON ? 1 : -1) > 0)
			plotter->color(0, 65535, 0);
		else
			plotter->color(65535, 0, 0);
		if (spin1y == spin2y && spin1x + 1 < spin2x) {
			plotter->fline(spin1x, spin1y, spin1x - 0.5, spin1y);
			plotter->fline(spin2x, spin2y, spin2x + 0.5, spin2y);
		} else if (spin1x == spin2x && spin2y + 1 < spin1y) {
			plotter->fline(spin1x, spin1y, spin1x, spin1y - 0.5);
			plotter->fline(spin2x, spin2y, spin2x, spin2y + 0.5);
		} else
			plotter->line(spin1x, spin1y, spin2x, spin2y);
	}
	plotter->filltype(1);
	for (size_t y = 0; y < height; y++)
		for (size_t x = 0; x < width; x++) {
			plotter->color(65535, 65535, 65535);
			plotter->fcircle(x, y, 0.25);
			if (c[x+y*width] == ON)
				plotter->color(20000, 20000, 20000);
			else
				plotter->color(50000, 50000, 50000);
			plotter->fcircle(x, y, 0.2);
		}
	plotter->closepl();
	delete plotter;
	out.close();
	return true;
}
#endif
