/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _DISCRETESTATISTICS_H_
#define _DISCRETESTATISTICS_H_


#include <algorithm>
#include <cmath>
#include <stdexcept>
#include <vector>
#include "util.h"


template <class T> class DiscreteStatistics {
public:
	DiscreteStatistics();

	size_t samples();
	T min() throw(std::domain_error);
	T max() throw(std::domain_error);
	double mean() throw(std::domain_error);
	double stddev() throw(std::domain_error);
	double median() throw(std::domain_error);

	size_t size();
	void add(T sample);
	void clear();

private:
	void ensureCached() throw(std::domain_error);

	bool cached;
	T _min, _max;
	double _mean, _stddev, _median;
	std::vector<T> data;
};


template <class T> inline DiscreteStatistics<T>::DiscreteStatistics()
: cached(false) {
}

template <class T> inline size_t DiscreteStatistics<T>::samples() {
	return data.size();
}

template <class T> inline T DiscreteStatistics<T>::min() throw(std::domain_error) {
	ensureCached();
	return _min;
}

template <class T> inline T DiscreteStatistics<T>::max() throw(std::domain_error) {
	ensureCached();
	return _max;
}

template <class T> inline double DiscreteStatistics<T>::mean() throw(std::domain_error) {
	ensureCached();
	return _mean;
}

template <class T> inline double DiscreteStatistics<T>::stddev() throw(std::domain_error) {
	ensureCached();
	return _stddev;
}

template <class T> inline double DiscreteStatistics<T>::median() throw(std::domain_error) {
	ensureCached();
	return _median;
}

template <class T> inline size_t DiscreteStatistics<T>::size() {
	return data.size();
}

template <class T> inline void DiscreteStatistics<T>::add(T sample) {
	cached = false;
	data.push_back(sample);
}

template <class T> inline void DiscreteStatistics<T>::clear() {
	data.clear();
	cached = false;
}

template <class T> void DiscreteStatistics<T>::ensureCached() throw(std::domain_error) {
	if (cached)
		return;
	if (data.size() < 2)
		throw std::domain_error("");
	sort(data.begin(), data.end());
	_min = data[0];
	_max = data[data.size()-1];
	_mean = 0;
	for (size_t i = 0; i < data.size(); i++)
		_mean += data[i];
	_mean /= data.size();
	_stddev = 0;
	for (size_t i = 0; i < data.size(); i++) {
		double tmp = _mean - data[i];
		_stddev += tmp * tmp;
	}
	_stddev = std::sqrt(_stddev / (data.size() - 1));
	if (data.size()%2 == 0)
		_median = (data[data.size()/2-1] + data[data.size()/2]) / 2.0;
	else
		_median = data[data.size()/2];
	cached = true;
}


typedef DiscreteStatistics<long> LongStatistics;
typedef DiscreteStatistics<double> DoubleStatistics;


#endif
