/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <new>
#include <sstream>
#include "Random.h"
#include "FiniteIntegerSet.h"


using namespace std;


FiniteIntegerSet::MemberIterator& FiniteIntegerSet::MemberIterator::operator++() throw(domain_error) {
	if (!hasNext())
		throw domain_error("FiniteIntegerSet::MemberIterator::operator++() domain_error: no next element");

	if (path.empty()) {
		Node *node = set->root;
		while (node != set->sentinel) {
			path.push(node);
			node = node->left;
		}
	} else {
		Node *node = path.top();
		if (node->right != set->sentinel) {
			node = node->right;
			while (node != set->sentinel) {
				path.push(node);
				node = node->left;
			}
		} else {
			path.pop();
			Node *parent = path.top();
			while (node == parent->right) {
				node = parent;
				path.pop();
				parent = path.top();
			}
		}
	}
	index++;
	return *this;
}

FiniteIntegerSet::MemberIterator& FiniteIntegerSet::MemberIterator::operator--() throw(domain_error) {
	if (!hasPrevious())
		throw domain_error("FiniteIntegerSet::MemberIterator::operator--() domain_error: no next element");

	if (path.empty()) {
		Node *node = set->root;
		while (node != set->sentinel) {
			path.push(node);
			node = node->right;
		}
		index = set->numberOfMembers();
	} else {
		Node *node = path.top();
		if (node->left != set->sentinel) {
			node = node->left;
			while (node != set->sentinel) {
				path.push(node);
				node = node->right;
			}
		} else {
			path.pop();
			Node *parent = path.top();
			while (node == parent->left) {
				node = parent;
				path.pop();
				parent = path.top();
			}
		}
		index--;
	}
	return *this;
}


FiniteIntegerSet::FiniteIntegerSet(unsigned int max)
: max(max) {
	sentinel = new Node();
	sentinel->color = BLACK;
	sentinel->left = sentinel->right = sentinel;
	root = sentinel;
}

FiniteIntegerSet::~FiniteIntegerSet() {
	deleteTree(root);
	delete sentinel;
}

bool FiniteIntegerSet::isMember(unsigned int value) const throw(out_of_range) {
	if (value > max) {
		stringstream ss;
		ss << "FiniteIntegerSet::isMember() out_of_range: Value " << value << " is too large";
		throw out_of_range(ss.str());
	}

	Node *node = root;
	while (node != sentinel) {
		if (value < node->value)
			node = node->left;
		else if (value > node->value)
			node = node->right;
		else
			return true;
	}
	return false;
}

unsigned int FiniteIntegerSet::randomMember() const throw(domain_error) {
	if (numberOfMembers() == 0)
		throw domain_error("FiniteIntegerSet::randomMember() domain_error: There are no member integers");

	unsigned int target = g_Random.next32(0, numberOfMembers() - 1);
	Node *node = root;
	unsigned int low = 0;
	while (true) {
		unsigned int membersLeft = node->value - node->nonLeft - low;
		if (target < membersLeft)
			node = node->left;
		else if (target > membersLeft) {
			target -= membersLeft + 1;
			low = node->value + 1;
			node = node->right;
		} else
			return node->value;
	}
	assert(false);
	return 0;
}

unsigned int FiniteIntegerSet::randomNonMember() const throw(domain_error) {
	if (numberOfNonMembers() == 0)
		throw domain_error("FiniteIntegerSet::randomNonMember() domain_error: There are no non-member integers");

	unsigned int target = g_Random.next32(0, numberOfNonMembers() - 1);
	if (root == sentinel)
		return target;

	Node *node = root;
	while (true) {
		if (target < node->nonLeft) {
			if (node->left == sentinel)
				return node->value - node->nonLeft + target;
			node = node->left;
		} else {
			if (node->right == sentinel)
				return node->value - node->nonLeft + target + 1;
			target -= node->nonLeft;
			node = node->right;
		}
	}
	assert(false);
	return 0;
}

void FiniteIntegerSet::addMember(unsigned int value) throw(out_of_range) {
	if (value > max) {
		stringstream ss;
		ss << "FiniteIntegerSet::addMember() out_of_range: Value " << value << " is too large";
		throw out_of_range(ss.str());
	}

	if (root == sentinel) {
		root = new Node();
		root->value = value;
		root->color = BLACK;
		root->left = root->right = sentinel;
		root->nonLeft = value;
		root->nonRight = max - value;
		return;
	}

	// Descend down the search path
	stack<Node*> path;
	Node *node = root;
	while (true) {
		if (value == node->value) {
			assert(false);
			stringstream ss;
			ss << "FiniteIntegerSet::addMember() out_of_range: Value " << value << " is already a member";
			throw out_of_range(ss.str());
		}
		path.push(node);
		if (value < node->value) {
			// Add on the left side
			assert(node->nonLeft > 0);
			node->nonLeft--;
			if (node->left != sentinel) {
				node = node->left;
				continue;
			}
			// Add a new left child
			Node *child = node->left = new Node();
			child->value = value;
			child->color = RED;
			child->left = child->right = sentinel;
			child->nonRight = node->value - value - 1;
			child->nonLeft = node->nonLeft - child->nonRight;
			break;
		} else {	
			// Add on the right side
			assert(node->nonRight > 0);
			node->nonRight--;
			if (node->right != sentinel) {
				node = node->right;
				continue;
			}
			// Add a new right child
			Node *child = node->right = new Node();
			child->value = value;
			child->color = RED;
			child->left = child->right = sentinel;
			child->nonLeft = value - node->value - 1;
			child->nonRight = node->nonRight - child->nonLeft;
			break;
		}
	}

	// Fix the tree on the way back up
	// TODO: add early termination condition
	while (!path.empty()) {
		node = path.top();
		path.pop();
		fixUp(node);
	}
	root->color = BLACK;
}

void FiniteIntegerSet::removeMember(unsigned int value) throw(out_of_range) {
	if (value > max) {
		stringstream ss;
		ss << "FiniteIntegerSet::removeMember() out_of_range: Value " << value << " is too large";
		throw out_of_range(ss.str());
	}

	// Temporarily violate root condition for the sake of flipping colors
	if (root->left->color == BLACK && root->right->color == BLACK)
		root->color = RED;

	// Descend down the search path
	stack<Node*> path;
	Node *node = root;
	while (true) {
		if (node == sentinel) {
			assert(false);
			stringstream ss;
			ss << "FiniteIntegerSet::removeMember() out_of_range: Value " << value << " is already not a member";
			throw out_of_range(ss.str());
		}
		if (value < node->value) {
			if (node->left != sentinel && node->left->color == BLACK && node->left->left->color == BLACK)
				moveRedLeft(node);
			path.push(node);
			node->nonLeft++;
			node = node->left;
		} else {
			if (node->left->color == RED)
				rotateRight(node);
			if (node->right != sentinel && node->right->color == BLACK && node->right->left->color == BLACK)
				moveRedRight(node);
			if (value == node->value) {
				if (node->left == sentinel && node->right == sentinel) {
					// The node is a leaf
					if (node == root) {
						root = sentinel;
						delete node;
						return;
					} else {
						assert(node->color == RED);
						Node *parent = path.top();
						if (node == parent->left)
							parent->left = sentinel;
						else
							parent->right = sentinel;
						delete node;
						break;
					}
				} else {
					// Swap target with successor
					Node *succ = successor(node);
					node->value = succ->value;
					succ->value = value;
	
					// Update non-members
					node->nonLeft += succ->nonLeft + 1;
					Node *tmp = node->left;
					while (tmp != sentinel) {
						tmp->nonRight += succ->nonLeft + 1;
						tmp = tmp->right;
					}
					node->nonRight -= succ->nonLeft + 1;
					tmp = node->right;
					while (tmp != sentinel) {
						tmp->nonLeft -= succ->nonLeft + 1;
						tmp = tmp->left;
					}
				}
			}
			path.push(node);
			node->nonRight++;
			node = node->right;
		}
	}

	// Fix the tree on the way back up
	while (!path.empty()) {
		node = path.top();
		path.pop();
		fixUp(node);
	}
	root->color = BLACK;
}

void FiniteIntegerSet::remax(unsigned int max) {
	this->max = max;
	deleteTree(root);
	root = sentinel;
}

void FiniteIntegerSet::deleteTree(Node* node) {
	if (node == sentinel)
		return;
	deleteTree(node->left);
	deleteTree(node->right);
}

void FiniteIntegerSet::rotateLeft(Node* node) {
//	assert(node->left->color == BLACK);
//	assert(node->right->color == RED);

	// Update links (child is moved from right to left)
	Node *child = node->right;
	node->right = child->right;
	child->right = child->left;
	child->left = node->left;
	node->left = child;

	// Swap values
	unsigned int tmpValue = node->value;
	node->value = child->value;
	child->value = tmpValue;

	// Update non-member values
	size_t tmpNon = child->nonLeft;
	child->nonLeft = node->nonLeft;
	node->nonLeft += tmpNon;
	node->nonRight = child->nonRight;
	child->nonRight = tmpNon;
} 

void FiniteIntegerSet::rotateRight(Node* node) {
//	assert(node->left->color == RED);
//	assert(node->right->color == BLACK);

	// Update links (child is moved from left to right)
	Node *child = node->left;
	node->left = child->left;
	child->left = child->right;
	child->right = node->right;
	node->right = child;

	// Swap values
	unsigned int tmpValue = node->value;
	node->value = child->value;
	child->value = tmpValue;

	// Update non-member values
	size_t tmpNon = child->nonRight;
	child->nonRight = node->nonRight;
	node->nonRight += tmpNon;
	node->nonLeft = child->nonLeft;
	child->nonLeft = tmpNon;
} 

inline void FiniteIntegerSet::flipColors(Node* node) {
	assert(node->color != node->left->color);
	assert(node->color != node->right->color);

	if (node->color == BLACK) {
		node->color = RED;
		node->left->color = node->right->color = BLACK;
	} else {
		node->color = BLACK;
		node->left->color = node->right->color = RED;
	}
}

inline void FiniteIntegerSet::fixUp(Node* node) {
	if (node->right->color == RED)
		rotateLeft(node);
	if (node->left->color == RED) {
		if (node->left->left->color == RED)
			rotateRight(node);
		if (node->right->color == RED) {
			assert(node->color == BLACK);
			flipColors(node);
		}
	}
}

inline void FiniteIntegerSet::moveRedRight(Node* node) {
	assert(node->color == RED);
	assert(node->left->color == BLACK);
	assert(node->right->color == BLACK);

	flipColors(node);
	if (node->left->left->color == RED) {
		rotateRight(node);
		flipColors(node);
	}
}

inline void FiniteIntegerSet::moveRedLeft(Node* node) {
	assert(node->color == RED);
	assert(node->left->color == BLACK);
	assert(node->right->color == BLACK);

	flipColors(node);
	if (node->right->left->color == RED) {
		rotateRight(node->right);
		rotateLeft(node);
		flipColors(node);
	}
}

inline FiniteIntegerSet::Node* FiniteIntegerSet::successor(Node* node) const {
	node = node->right;
	while (node->left != sentinel)
		node = node->left;
	return node;
}

#ifdef SELFTEST
size_t blackHeight, numMembers, numNonMembers;

bool FiniteIntegerSet::isValid() const {
	assert(root != NULL);
	assert(root->color == BLACK);
	assert(sentinel->color == BLACK);
	assert(sentinel->left == sentinel);
	assert(sentinel->right == sentinel);
	assert(sentinel->nonLeft == 0);
	assert(sentinel->nonRight == 0);

	if (root == sentinel)
		return true;
	blackHeight = numMembers = numNonMembers = 0;
	isValid(root, 0);
	assert(numberOfMembers() == numMembers);
	assert(numberOfNonMembers() == numNonMembers);
	assert(numMembers + numNonMembers == max + 1);

	return true;
}

void FiniteIntegerSet::isValid(Node* node, size_t blackDepth) const {
	if (node->color == BLACK)
		blackDepth++;
	if (node == sentinel)
		if (blackHeight == 0)
			blackHeight = blackDepth;
		else {
			assert(blackHeight == blackDepth);
		}
	else {
		numMembers++;
		assert(node->color == BLACK || (node->left->color == BLACK && node->right->color == BLACK));
		assert(node->left->color == RED || node->right->color == BLACK);
		if (node->left != sentinel) {
			assert(node->nonLeft == node->left->nonLeft + node->left->nonRight);
			assert(node->value > node->left->value);
			assert(node->left->nonRight < node->value - node->left->value);
		} else {
			assert(node->nonLeft <= node->value);
			numNonMembers += node->nonLeft;
		}
		if (node->right != sentinel) {
			assert(node->nonRight == node->right->nonLeft + node->right->nonRight);
			assert(node->value < node->right->value);
			assert(node->right->nonLeft < node->right->value - node->value);
		} else {
			assert(node->nonRight <= max - node->value);
			numNonMembers += node->nonRight;
		}
		isValid(node->left, blackDepth);
		isValid(node->right, blackDepth);
	}
}

void FiniteIntegerSet::selfTest() {
	FiniteIntegerSet set(9);

	assert(set.isValid());
	assert(set.numberOfMembers() == 0);
	assert(set.numberOfNonMembers() == 10);

	bool exceptionThrown = false;
	try {
		set.addMember(10);
	} catch (out_of_range& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);
	exceptionThrown = false;
	try {
		set.isMember(10);
	} catch (out_of_range& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);
	exceptionThrown = false;
	try {
		set.randomMember();
	} catch (domain_error& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);
	
	set.addMember(5);
	assert(set.isValid());
	assert(set.numberOfMembers() == 1);
	assert(set.numberOfNonMembers() == 9);
	assert(set.isMember(5));
	assert(!set.isMember(3));
	assert(!set.isMember(8));

	set.addMember(3);
	assert(set.isValid());
	assert(set.numberOfMembers() == 2);
	assert(set.numberOfNonMembers() == 8);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(8));

	set.addMember(8);
	assert(set.isValid());
	assert(set.numberOfMembers() == 3);
	assert(set.numberOfNonMembers() == 7);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(set.isMember(8));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(6));
	assert(!set.isMember(9));

	set.clear();
	assert(set.isValid());
	assert(set.numberOfMembers() == 0);
	assert(set.numberOfNonMembers() == 10);
	assert(!set.isMember(3));
	assert(!set.isMember(5));
	assert(!set.isMember(8));

	set.addMember(5);
	set.addMember(8);
	set.addMember(3);
	assert(set.isValid());
	assert(set.numberOfMembers() == 3);
	assert(set.numberOfNonMembers() == 7);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(set.isMember(8));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(6));
	assert(!set.isMember(9));

	set.clear();
	set.addMember(3);
	set.addMember(5);
	set.addMember(8);
	assert(set.isValid());
	assert(set.numberOfMembers() == 3);
	assert(set.numberOfNonMembers() == 7);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(set.isMember(8));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(6));
	assert(!set.isMember(9));

	set.clear();
	set.addMember(8);
	set.addMember(5);
	set.addMember(3);
	assert(set.isValid());
	assert(set.numberOfMembers() == 3);
	assert(set.numberOfNonMembers() == 7);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(set.isMember(8));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(6));
	assert(!set.isMember(9));

	set.clear();
	set.addMember(8);
	set.addMember(3);
	set.addMember(5);
	assert(set.isValid());
	assert(set.numberOfMembers() == 3);
	assert(set.numberOfNonMembers() == 7);
	assert(set.isMember(3));
	assert(set.isMember(5));
	assert(set.isMember(8));
	assert(!set.isMember(2));
	assert(!set.isMember(4));
	assert(!set.isMember(6));
	assert(!set.isMember(9));

	set.clear();
	for (int i = 0; i <= 9; i++) {
		set.addMember(i);
		assert(set.isValid());		
		assert(set.isMember(i));
	}
	assert(set.numberOfMembers() == 10);
	assert(set.numberOfNonMembers() == 0);

	set.clear();
	for (int i = 9; i >= 0; i--) {
		set.addMember(i);
		assert(set.isValid());
		assert(set.isMember(i));
	}
	assert(set.numberOfMembers() == 10);
	assert(set.numberOfNonMembers() == 0);

	FiniteIntegerSet set2(99);
	for (int i = 0; i <= 9; i++)
		set2.addMember(i);
	assert(set2.isValid());
	assert(set2.numberOfMembers() == 10);
	assert(set2.numberOfNonMembers() == 90);
	assert(set2.isMember(5));
	assert(!set2.isMember(50));

	for (int i = 10; i <= 99; i++) {
		set2.addMember(i);
		assert(set2.isValid());	
	}
	assert(set2.numberOfMembers() == 100);
	assert(set2.numberOfNonMembers() == 0);

	set2.clear();
	for (int i = 99; i >= 0; i -= 2) {
		set2.addMember(i);
		assert(set2.isValid());
	}
	assert(set2.numberOfMembers() == 50);
	assert(set2.numberOfNonMembers() == 50);
	for (int i = 0; i <= 99; i++)
		if (i % 2) {
			assert(set2.isMember(i));
		} else {
			assert(!set2.isMember(i));
		}
	for (int i = 0; i < 1000; i++)
		assert(set2.isMember(set2.randomMember()));
	for (int i = 0; i < 1000; i++)
		assert(!set2.isMember(set2.randomNonMember()));
	for (unsigned int i = 0; i <= 99; i++)
		if (i % 2)
			while (set2.randomMember() != i);
		else
			while (set2.randomNonMember() != i);

	for (int i = 0; i <= 99; i += 2) {
		set2.addMember(i);
		assert(set2.isValid());
	}
	assert(set2.numberOfMembers() == 100);
	assert(set2.numberOfNonMembers() == 0);

	set2.clear();
	for (int i = 0; i < 50; i++)
		set2.addMember(set2.randomNonMember());
	for (unsigned int i = 0; i <= 99; i++)
		if (set2.isMember(i))
			while (set2.randomMember() != i);
		else
			while (set2.randomNonMember() != i);

	while (set2.numberOfNonMembers() > 0)
		set2.addMember(set2.randomNonMember());
	assert(set2.numberOfMembers() == 100);
	assert(set2.numberOfNonMembers() == 0);

	for (int i = 99; i >= 0; i -= 2) {
		set2.removeMember(i);
		assert(set2.isValid());
	}
	for (int i = 0; i <= 99; i += 2) {
		set2.removeMember(i);
		assert(set2.isValid());
	}

	while (set2.numberOfNonMembers() > 0)
		set2.addMember(set2.randomNonMember());
	assert(set2.isValid());
	while (set2.numberOfMembers() > 0) {
		set2.removeMember(set2.randomMember());
		assert(set2.isValid());
	}
	while (set2.numberOfNonMembers() > 0)
		set2.addMember(set2.randomNonMember());
	assert(set2.isValid());
	while (set2.numberOfMembers() > 0) {
		set2.removeMember(set2.randomMember());
		assert(set2.isValid());
	}

	FiniteIntegerSet::MemberIterator it = set2.iterator();
	assert(!it.hasNext());
	assert(!it.hasPrevious());

	for (int i = 0; i <= 9; i++)
		set2.addMember(i);
	it = set2.iterator();
	++it;
	assert(it.hasNext());
	assert(!it.hasPrevious());
	assert(*it == 0);
	for (unsigned int i = 1; i <= 8; i++) {
		++it;
		assert(it.hasNext());
		assert(it.hasPrevious());
		assert(*it == i);
	}
	++it;
	assert(!it.hasNext());
	assert(it.hasPrevious());
	assert(*it == 9);

	exceptionThrown = false;
	try {
		++it;
	} catch (domain_error& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);

	it = set2.iterator();
	--it;
	assert(!it.hasNext());
	assert(it.hasPrevious());
	assert(*it == 9);
	for (unsigned int i = 8; i >= 1; i--) {
		--it;
		assert(it.hasNext());
		assert(it.hasPrevious());
		assert(*it == i);
	}
	--it;
	assert(it.hasNext());
	assert(!it.hasPrevious());
	assert(*it == 0);

	exceptionThrown = false;
	try {
		--it;
	} catch (domain_error& e) {
		exceptionThrown = true;
	}
	assert(exceptionThrown);
}

#endif
