/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#ifndef _FINITEINTEGERSET_H_
#define _FINITEINTEGERSET_H_


#include <cstddef>
#include <stack>
#include <stdexcept>


//
// This class represents a set of integers within the range [0, max].  The
// members of the set are stored in a left-leaning red-black tree (Sedgewick,
// 2008).
//

class FiniteIntegerSet {
private:
	struct Node;

public:
	friend class MemberIterator;
	class MemberIterator {
		friend class FiniteIntegerSet;

	public:
		MemberIterator();

		bool hasNext() const throw(std::domain_error);
		bool hasPrevious() const throw(std::domain_error);
		bool operator==(const MemberIterator& it) const;
		bool operator!=(const MemberIterator& it) const;
		unsigned int operator*() const throw(std::domain_error);

		MemberIterator& operator++() throw(std::domain_error); // pre
		MemberIterator& operator--() throw(std::domain_error); // pre

	private:
		MemberIterator(const FiniteIntegerSet* set);

		const FiniteIntegerSet *set;
		std::stack<Node*> path;
		unsigned int index;
	};


	FiniteIntegerSet(unsigned int max);
	~FiniteIntegerSet();

	size_t numberOfMembers() const;
	size_t numberOfNonMembers() const;
	bool isMember(unsigned int value) const throw(std::out_of_range);
	unsigned int randomMember() const throw(std::domain_error);
	unsigned int randomNonMember() const throw(std::domain_error);
	// Iterators are invalidated by changes to the set
	MemberIterator iterator() const;

	void addMember(unsigned int value) throw(std::out_of_range);
	void removeMember(unsigned int value) throw(std::out_of_range);
	void clear();
	void remax(unsigned int max);

#ifdef SELFTEST
	bool isValid() const;
	static void selfTest();
#endif

private:
	enum Color {
		RED, BLACK
	};
	struct Node {
		unsigned int value;
		enum Color color;
		Node *left, *right;
		size_t nonLeft, nonRight;
	};

	void deleteTree(Node* node);
	static void rotateLeft(Node* node);
	static void rotateRight(Node* node);
	static void flipColors(Node* node);
	static void fixUp(Node* node);
	static void moveRedRight(Node* node);
	static void moveRedLeft(Node* node);
	Node* successor(Node* node) const;

#ifdef SELFTEST
	void isValid(Node* node, size_t blackDepth) const;
#endif

	unsigned int max;
	Node *root;
	Node *sentinel;
};


inline FiniteIntegerSet::MemberIterator::MemberIterator()
: set(NULL), index(0) {
}

inline bool FiniteIntegerSet::MemberIterator::hasNext() const throw(std::domain_error) {
	if (set == NULL)
		throw std::domain_error("FiniteIntegerSet::MemberIterator::hasNext() domain_error: not a valid iterator");

	return index < set->numberOfMembers();
}

inline bool FiniteIntegerSet::MemberIterator::hasPrevious() const throw(std::domain_error) {
	if (set == NULL)
		throw std::domain_error("FiniteIntegerSet::MemberIterator::hasNext() domain_error: not a valid iterator");

	return index != 1 && set->numberOfMembers() > 0;
}

inline bool FiniteIntegerSet::MemberIterator::operator==(const MemberIterator& it) const {
	return set == it.set && (set == NULL || index == it.index);
}

inline bool FiniteIntegerSet::MemberIterator::operator!=(const MemberIterator& it) const {
	return !(*this == it);
}

inline unsigned int FiniteIntegerSet::MemberIterator::operator*() const throw(std::domain_error) {
	if (path.empty())
		throw std::domain_error("FiniteIntegerSet::MemberIterator::operator*() domain_error: no valid element");
	return path.top()->value;
}

inline FiniteIntegerSet::MemberIterator::MemberIterator(const FiniteIntegerSet* set)
: set(set), index(0) { 
}


inline size_t FiniteIntegerSet::numberOfMembers() const {
	if (root == sentinel)
		return 0;
	return max + 1 - root->nonLeft - root->nonRight;
}

inline size_t FiniteIntegerSet::numberOfNonMembers() const {
	if (root == sentinel)
		return max + 1;
	return root->nonLeft + root->nonRight;
}

inline FiniteIntegerSet::MemberIterator FiniteIntegerSet::iterator() const {
	return MemberIterator(this);
}

inline void FiniteIntegerSet::clear() {
	deleteTree(root);
	root = sentinel;
}


#endif

