/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <new>
#include <sstream>
#include "util.h"
#include "Multitype.h"


using namespace std;


#define THROW_EXPECTED(x) {assert(false); throw invalid_argument("Multitype expected an argument of type x");}
#define THROW_UNKNOWN {assert(false); throw invalid_argument("Cannot perform operation involving an undefined Multitype.");}
#define THROW_BAD_CAST {assert(false); throw bad_cast();}


Multitype::Multitype()
: type(UNKNOWN), as_string(NULL) {
}

Multitype::Multitype(long l)
: type(LONG), as_long(l), as_string(NULL) {
}

Multitype::Multitype(double d)
: type(DOUBLE), as_double(d), as_string(NULL) {
}

Multitype::Multitype(const char* s)
: type(STRING), as_string(new string(s)) {
}

Multitype::Multitype(const string& s)
: type(STRING), as_string(new string(s)) {
}

Multitype::Multitype(const Multitype& m)
: type(m.type), as_string(NULL) {
	switch(type) {
	case LONG:	as_long = m.as_long; break;
	case DOUBLE:	as_double = m.as_double; break;
	case STRING:	as_string = new string(*m.as_string); break;
	case UNKNOWN:	;
	}
}

Multitype::~Multitype() {
	delete as_string;
}

Multitype::TYPE Multitype::getType() const {
	return type;
}

Multitype& Multitype::operator=(long l) throw(invalid_argument) {
	switch (type) {
		case UNKNOWN:	type = LONG;
		case LONG:	as_long = l; break;
		case DOUBLE:	THROW_EXPECTED(DOUBLE);
		case STRING:	THROW_EXPECTED(STRING);
	}
	return *this;
}

Multitype& Multitype::operator=(double d) throw(invalid_argument) {
	switch (type) {
		case UNKNOWN:	type = DOUBLE;
		case DOUBLE:	as_double = d; break;
		case LONG:	THROW_EXPECTED(LONG);
		case STRING:	THROW_EXPECTED(STRING);
	}
	return *this;
}

Multitype& Multitype::operator=(const char* s) throw(invalid_argument) {
	switch (type) {
		case UNKNOWN:	type = STRING;
		             	if (as_string == NULL) {
		             		as_string = new string(s);
		             		break;
		             	}
		case STRING:	*as_string = s; break;
		case LONG:	THROW_EXPECTED(LONG);
		case DOUBLE:	THROW_EXPECTED(DOUBLE);
	}
	return *this;
}

Multitype& Multitype::operator=(const string& s) throw(invalid_argument) {
	switch (type) {
		case UNKNOWN:	type = STRING;
		             	if (as_string == NULL) {
		             		as_string = new string(s);
		             		break;
		             	}
		case STRING:	*as_string = s; break;
		case LONG:	THROW_EXPECTED(LONG);
		case DOUBLE:	THROW_EXPECTED(DOUBLE);
	}
	return *this;
}

Multitype& Multitype::operator=(const Multitype& m) throw(invalid_argument) {
	if (this == &m)
		return *this;
	if (type == UNKNOWN) {
		type = m.type;
		if (type == STRING && as_string == NULL) {
			as_string = new string(*m.as_string);
			return *this;
		}
	} else if (type != m.type) {
		switch (type) {
			case LONG:	THROW_EXPECTED(LONG);
			case DOUBLE:	THROW_EXPECTED(DOUBLE);
			case STRING:	THROW_EXPECTED(STRING);
			case UNKNOWN:	assert(false);
		}
	}
	switch (type) {
		case LONG:	as_long = m.as_long; break;
		case DOUBLE:	as_double = m.as_double; break;
		case STRING:	*as_string = *m.as_string; break;
		case UNKNOWN:	THROW_UNKNOWN;
	}
	return *this;
}

void Multitype::markUnknown() {
	type = UNKNOWN;
}

Multitype::operator const long&() const throw(bad_cast) {
	if (type == UNKNOWN)
		const_cast<Multitype*>(this)->parseLong();
	if (type != LONG)
		THROW_BAD_CAST;
	return as_long;
}

Multitype::operator long&() throw(bad_cast) {
	if (type == LONG)
		return as_long;
	else if (type == UNKNOWN) {
		if (!parseLong())
			THROW_BAD_CAST;
		return as_long;
	} else
		THROW_BAD_CAST;
}

Multitype::operator const double&() const throw(bad_cast) {
	if (type == UNKNOWN)
		const_cast<Multitype*>(this)->parseDouble();
	if (type != DOUBLE)
		THROW_BAD_CAST;
	return as_double;
}

Multitype::operator double&() throw(bad_cast) {
	if (type == DOUBLE)
		return as_double;
	else if (type == UNKNOWN) {
		if (!parseDouble())
			THROW_BAD_CAST;
		return as_double;
	} else
		THROW_BAD_CAST;
}

Multitype::operator const string&() const throw(bad_cast) {
	if (type == UNKNOWN && as_string != NULL)
		const_cast<Multitype*>(this)->type = STRING;
	if (type != STRING)
		THROW_BAD_CAST;
	return *as_string;
}

Multitype::operator string&() throw(bad_cast) {
	if (type == STRING)
		return *as_string;
	else if (type == UNKNOWN) {
		if (as_string == NULL)
			as_string = new string("");
//			THROW_BAD_CAST;
		type = STRING;
		return *as_string;
	} else
		THROW_BAD_CAST;
}

ostream& operator<<(ostream& out, const Multitype& m) {
	switch (m.type) {
	case Multitype::LONG: out << m.as_long; break;
	case Multitype::DOUBLE: out << m.as_double; break;
	case Multitype::UNKNOWN:
	case Multitype::STRING: out << *m.as_string; break;
	}
	return out;
}

bool Multitype::parseLong() {
	if (as_string == NULL) {
		type = LONG;
		as_long = 0;
		return true;
	}

	stringstream ss(*as_string);
	long ltmp;
	ss >> ltmp >> detecteof;
	if (!ss.eof())
		return false;
	type = LONG;
	as_long = ltmp;
	return true;
}

bool Multitype::parseDouble() {
	if (as_string == NULL) {
		type = DOUBLE;
		as_double = 0.0;
		return true;
	}

	stringstream ss(*as_string);
	double dtmp;
	ss >> dtmp >> detecteof;
	if (!ss.eof())
		return false;
	type = DOUBLE;
	as_double = dtmp;
	return true;
}
