/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <cctype>
#include <sstream>
#include "ParameterFile.h"


using namespace std;


ParameterFile::ParameterFile() {
}

ParameterFile::ParameterFile(const ParameterFile& pf) 
: sets(pf.sets) {
}

ParameterFile::ParameterFile(const char* filename) {
	assert(filename != NULL);
	load(filename);
}

size_t ParameterFile::size() const {
	return sets.size();
}

bool ParameterFile::contains(const std::string& section) const {
	return sets.find(section) != sets.end();
}

ParameterFile::iterator ParameterFile::find(const string& section) {
	return sets.find(section);
}

ParameterFile::const_iterator ParameterFile::find(const string& section) const {
	return sets.find(section);
}

ParameterSet& ParameterFile::operator[](const string& section) {
	iterator it = sets.find(section);
	if (it != sets.end())
		return it->second;
	return sets.insert(make_pair(section, ParameterSet())).first->second;
}

const ParameterSet& ParameterFile::operator[](const string& section) const throw(invalid_argument) {
	const_iterator it = find(section);
	if (it == sets.end()) {
		assert(false);
		stringstream ss;
		ss << "Section '" << section << "' doesn't exist.";
		throw invalid_argument(ss.str());
	}
	return it->second;
}

bool ParameterFile::get(const string& section, ParameterSet& dest) const {
	const_iterator it = find(section);
	if (it == sets.end()) 
		return false;
	dest.add(it->second);
	return true;
}

ParameterFile::iterator ParameterFile::begin() {
	return sets.begin();
}

ParameterFile::const_iterator ParameterFile::begin() const {
	return sets.begin();
}

ParameterFile::iterator ParameterFile::end() {
	return sets.end();
}

ParameterFile::const_iterator ParameterFile::end() const {
	return sets.end();
}

void ParameterFile::remove(const string& section) {
	sets.erase(section);
}

void ParameterFile::clear() {
	sets.clear();
}

istream& operator>>(istream& in, ParameterFile& pf) {
	string line;

	in >> detecteof;
	while (in.good()) {
		if (in.peek() == '#') {
			in >> endl;
			continue;
		}
		if (in.peek() != '[')
			break;
		in.get();	// eat '['
		getline(in, line);
		trim(line);	// cut off DOS linefeed
		if (line.size() > 0 && line[line.size()-1] == ']') // eat ']'
			line.erase(line.size()-1, 1);
		in >> pf[line] >> detecteof;
	}

	return in;
}

ostream& operator<<(ostream& out, const ParameterFile& pf) {
	ParameterFile::const_iterator it = pf.begin();
	while (it != pf.sets.end()) {
		out << '[' << it->first<< ']' << endl << it->second << endl;
		it++;
	}
	return out;
}
