/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <cctype>
#include <sstream>
#include "ParameterSet.h"


using namespace std;


ParameterSet::ParameterSet() {
}

ParameterSet::ParameterSet(const ParameterSet& ps)
: values(ps.values) {
}

ParameterSet::ParameterSet(const char* filename) {
	assert(filename != NULL);
	load(filename);
}

size_t ParameterSet::size() const {
	return values.size();
}

bool ParameterSet::contains(const std::string& name) const {
	return values.find(name) != values.end();
}

ParameterSet::iterator ParameterSet::find(const string& name) {
	return values.find(name);
}

ParameterSet::const_iterator ParameterSet::find(const string& name) const {
	return values.find(name);
}

Multitype& ParameterSet::operator[](const string& name) {
	iterator it = values.find(name);
	if (it != values.end())
		return it->second;
	return values.insert(make_pair(name, Multitype())).first->second;
}

const Multitype& ParameterSet::operator[](const string& name) const throw(invalid_argument) {
	const_iterator it = values.find(name);
	if (it == values.end()) {
		assert(false);
		stringstream ss;
		ss << "Parameter '" << name << "' doesn't exist.";
		throw invalid_argument(ss.str());
	}
	return it->second;
}

bool ParameterSet::get(const string& name, long& dest) const {
	const_iterator it = values.find(name);
	if (it == values.end())
		return false;
	dest = it->second;
	return true;
}

bool ParameterSet::get(const string& name, double& dest) const {
	const_iterator it = values.find(name);
	if (it == values.end())
		return false;
	dest = it->second;
	return true;
}

bool ParameterSet::get(const string& name, string& dest) const {
	const_iterator it = values.find(name);
	if (it == values.end())
		return false;
	dest = static_cast<const string&>(it->second);
	return true;
}

ParameterSet::iterator ParameterSet::begin() {
	return values.begin();
}

ParameterSet::const_iterator ParameterSet::begin() const {
	return values.begin();
}

ParameterSet::iterator ParameterSet::end() {
	return values.end();
}

ParameterSet::const_iterator ParameterSet::end() const {
	return values.end();
}

void ParameterSet::remove(const string& name) {
	values.erase(name);
}

void ParameterSet::clear() {
	values.clear();
}

void ParameterSet::add(const ParameterSet& ps) {
	values.insert(ps.begin(), ps.end());
}

istream& operator>>(istream& in, ParameterSet& ps) {
	string name, value;
	stringstream ss;

	in >> detecteof;
	while (in.good()) {
		ss.clear();
		ss.str("");
		getline(in, ss);
		ss >> ws;
		if (ss.eof())
			break;
		if (ss.peek() == '#') {
			in >> detecteof;
			continue;
		}
		ss >> name >> ws;
		value = "";
		getline(ss, value);
		trim(value);
		if (value.length() >= 2 && value[0] == '"' && value[value.length()-1] == '"') {
			value.erase(value.length()-1,1);
			value.erase(0,1);
		}
		(ps[name]=value).markUnknown();
		in >> detecteof;
	}

	return in;
}

ostream& operator<<(ostream& out, const ParameterSet& ps) {
	StreamFormat fmt;
	out << fmt;

	ParameterSet::const_iterator it = ps.begin();
	while (it != ps.values.end()) {
		out << it->first << " ";
		switch (it->second.getType()) {
		case Multitype::LONG:
		case Multitype::DOUBLE:
			out << it->second << endl; break;
		case Multitype::UNKNOWN:
		case Multitype::STRING:
			out << '"' << it->second << '"' << endl; break;
		}
		it++;
	}

	out << fmt;
	return out;
}
