/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cstring>
#include <new>
#include <string>
#include <fstream>
#include "MARLEDA.h"
#include "Random.h"
#include "Chi2Approx.h"
#include "OneMax.h"
#include "TrapMax.h"
#include "Rosenbrock.h"
#include "SpinGlass.h"
#include "Uniform.h"


using namespace std;


EvolutionTask* selectTask(const string& name);
void showUsage();


int main(int argc, char *argv[]) {
	if (argc == 1) {
		showUsage();
		return 0;
	}

	ParameterFile pf;
	ParameterSet ps;

	// Load all parameters
	for (int i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			if (i+1 >= argc) {
				cerr << "Error: missing argument to " << argv[i] << endl;
				return 1;
			}
			if (argv[i][1] == '-') {
				// remove parameter
				ps.remove(argv[++i]);
			} else {
				// parameter override
				(ps[argv[i]+1] = argv[i+1]).markUnknown();
				i++;
			}
		} else {
			// parameter file
			char *s = strchr(argv[i], ':');
			if (s == NULL) {
				// a parameter set file
				if (!ps.load(argv[i])) {
					cerr << "Error loading parameters from '" << argv[i] << "'" << endl;
					return 1;
				}
			} else if (s == argv[i]) {
				if (pf.size() == 0) {
					cerr << ":[section] notation requires previously specified parameter file" << endl;
					return 1;
				}
				if (!pf.get(s+1, ps)) {
					cerr << "Error loading parameters from '" << argv[i] << "'" << endl;
					return 1;
				}
			} else {
				*s = 0;
				if (!pf.load(argv[i])) {
					cerr << "Error loading parameter from '" << argv[i] << "'" << endl;
					return 1;
				}
				if (!pf.get(s+1, ps)) {
					*s = ':';
					cerr << "Error loading parameters from '" << argv[i] << "'" << endl;
					return 1;
				}
			}
		}
	}

	if (ps.contains("rngSeed"))
		g_Random.reseed(static_cast<long>(ps["rngSeed"]));

	string s;
	if (!ps.get("task", s)) {
		cerr << "Error: No evolution task specified" << endl;
		return 1;
	}
	EvolutionTask *task = selectTask(s);
	if (task == NULL) {
		cerr << "Error: Unable to create evolution task" << endl;
		return 1;
	}

	if (!g_Chi2Approx.load("chi2.dat"))
		throw runtime_error("Error: Unable to load chi2.dat");

	MARLEDA marleda;
	if (!marleda.init(ps, task))
		return 1;
	try {
		marleda.evolve();
	} catch (exception& e) {
		cerr << e.what() << endl;
		assert(false);
		return 1;
	}

	return 0;
}

EvolutionTask* selectTask(const string& name) {
	if (name == "onemax")
		return new OneMax();
	else if (name == "trapmax")
		return new TrapMax();
	else if (name == "rosenbrock")
		return new Rosenbrock();
	else if (name == "spin")
		return new SpinGlass();
	else if (name == "uniform")
		return new Uniform();
	cerr << "Error: unknown task '" << name << "'" << endl;
	return NULL;
}

void showUsage() {
	cout << "Usage: marleda [<FILE> | <PARAMETER>]..." << endl;
	cout << "   <FILE>s and <PARAMETER>s are evaluated in the order they are specified." << endl;
	cout << "   <FILE> - A parameter file of the form:" << endl;
	cout << "      <filename>             Reads parameters from a file containing a single" << endl;
	cout << "                             unnamed set of parameters." << endl;
	cout << "      <filename>:<section>   Reads a single set of parameters from a file" << endl;
	cout << "                             containing multiple named sets of parameters." << endl;
	cout << "      :<section>             Reads another single named set of parameters from" << endl;
	cout << "                             the most recently used parameter file." << endl;
	cout << "   <PARAMETER> - A parameter of either form:" << endl;
	cout << "      -<name> <value>" << endl;
	cout << "      --remove <name>" << endl << endl;
}
