/***********************************************************************
 * Copyright 2009, Matthew Alden
 *
 * This file is part of MARLEDA.
 *
 * MARLEDA is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MARLEDA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MARLEDA.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************/

#include <cassert>
#include <cmath>
#include <cstring>
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#include "util.h"


using namespace std;


void showUsage();
int trapmax(int argc, char* argv[]);
int spin(int argc, char* argv[]);


int main(int argc, char* argv[]) {
	if (argc < 2 || strcmp(argv[1], "--help") == 0) {
		showUsage();
		return 0;
	}

	if (strcmp(argv[1], "-trapmax") == 0)
		return trapmax(argc - 1, &argv[1]);
	if (strcmp(argv[1], "-spin") == 0)
		return spin(argc - 1, &argv[1]);
	showUsage();
	return 1;
}

void showUsage() {
	cout << "Usage: makemodel -trapmax <genes> <trapsize> <output-file>" << endl;
	cout << "       makemodel -spin <width> <height> <order> <output-file>" << endl;
}

int trapmax(int argc, char* argv[]) {
	if (argc != 4) {
		showUsage();
		return 1;
	}

	stringstream ss;
	ss << argv[1];
	int genes = -1;
	ss >> genes;
	if (genes < 2) {
		cerr << "Error: invalid number of genes '" << argv[1] << "'" << endl;
		return 1;
	}
	ss.clear();
	ss.str("");
	ss << argv[2];
	int trapsize = -1;
	ss >> trapsize;
	if (trapsize < 2) {
		cerr << "Error: invalid trapsize '" << argv[2] << "'" << endl;
		return 1;
	}
	if (genes%trapsize != 0) {
		cerr << "Error: genes must be a multiple of trapsize" << endl;
		return 1;
	}

	ofstream out(argv[3]);
	if (!out || !out.is_open()) {
		cerr << "Error: unable to open '" << argv[3] << "'" << endl;
		return 1;
	}

	out << genes << endl;
	for (size_t i = 0; i < genes; i++) {
		int block = i / trapsize;
		int low = block * trapsize;
		int high = low + trapsize - 1;
		out << trapsize - 1;
		for (size_t j = low; j <= high; j++) {
			if (j == i)
				continue;
			out << ' ' << j;
		}
		for (size_t j = 0; j < low; j++)
			out << ' ' << j;
		for (size_t j = high + 1; j < genes; j++)
			out << ' ' << j;
		out << endl;
	}
	out.close();

	return 0;
}

int spin(int argc, char* argv[]) {
	if (argc != 5) {
		showUsage();
		return 1;
	}

	stringstream ss;
	ss << argv[1];
	int width = -1;
	ss >> width;
	if (width < 2) {
		cerr << "Error: width must be at least 2" << endl;
		return 1;
	}
	ss.clear();
	ss.str("");
	ss << argv[2];
	int height = -1;
	ss >> height;
	if (height < 2) {
		cerr << "Error: height must be at least 2" << endl;
		return 1;
	}
	ss.clear();
	ss.str("");
	ss << argv[3];
	int order = -1;
	ss >> order;
	if (order < 1) {
		cerr << "Error: order must be at least 1" << endl;
		return 1;
	}

	ofstream out(argv[4]);
	if (!out || !out.is_open()) {
		cerr << "Error: unable to open '" << argv[4] << "'" << endl;
		return 1;
	}

	out << width * height << endl;
	vector<int> neighbors, nonneighbors;
	for (int y = 0; y < height; y++)
		for (int x = 0; x < width; x++) {
			neighbors.clear();
			nonneighbors.clear();
			for (int t = 0; t < height; t++) {
				int yDiff = abs(t - y);
				if (height - yDiff < yDiff)
					yDiff = height - yDiff;
				for (int s = 0; s < width; s++) {
					if (s == x && t == y)
						continue;
					int xDiff = abs(s - x);
					if (width - xDiff < xDiff)
						xDiff = width - xDiff;
					if (xDiff + yDiff <= order)
						neighbors.push_back(t * width + s);
					else
						nonneighbors.push_back(t * width + s);
				}
			}
			assert(neighbors.size() + nonneighbors.size() == width * height - 1);
			out << neighbors.size();
			for (size_t i = 0; i < neighbors.size(); i++)
				out << ' ' << neighbors[i];
			for (size_t i = 0; i < nonneighbors.size(); i++)
				out << ' ' << nonneighbors[i];
			out << endl;
		}
	out.close();

	return 0;
}
